/**********************************
  Aggregate Rate Controller (ARC)
  Author: Jae Chung
  Email:  goos@cs.wpi.edu
  Date:   Sep. 2003
***********************************/

#ifndef ns_arcq_h
#define ns_arcq_h

#include "ip.h"
#include "queue.h"
#include "config.h"
#include "timer-handler.h"

class ARCQueue;
class LinkDelay;

/*
 * ARC parameters, supplied by user
 */
struct arcp {
    int setbit;             /* true to set ECN congestion indication bit */
    int qib;                /* true if queue in bytes, false if packets */
    int bytes;              /* true if bytes used  (advantage to small pkts) */
    int mean_pktsize;       /* mean packet size (bytes) */

    double d;               /* statistics update interval (in sec) */
    double alpha;           /* integral control paramter */
    double tload;           /* target load (0<TL<=1) */
    int    tqs;             /* target queue size (in pkts) */
};


/*
 * ARC variables, maintained by ARC
 */
struct arcv {
    double sr;            /* service rate (in bytes per second) */
    TracedDouble esr;     /* estimated service rate (in bps) */    
    TracedDouble notif_p; /* congestion notification probability */

    int b_svcd;           /* bytes serviced */

    arcv() : sr(0.0), esr(0.0), notif_p(0.0), b_svcd(0) {}
};



/*
 * ARC Congestion Notification Probability Update Timer
 */
class ARCUpdateTimer : public TimerHandler {
public:
    ARCUpdateTimer(ARCQueue* t) : TimerHandler(), t_(t) {}
    inline virtual void expire(Event*);
 protected:
    ARCQueue* t_;
};


/*
 * ARC Queue
 */
class ARCQueue : public Queue {
 public:	
    ARCQueue();
    void update_stats();
    
 protected:
    int     command(int argc, const char*const* argv);
    void    reset();
    int     notify_early(double notif_p);
    void    enque(Packet* pkt);
    Packet* deque();
    void    trace(TracedVar*);	/* routine to write trace records */
    void ARCQueue::print_arcp();
    void ARCQueue::print_arcv();
    
    LinkDelay*     link_;	 /* outgoing link */
    PacketQueue    *q_; 	 /* underlying (usually) FIFO queue */
    NsObject*      de_drop_;     /* drop_early target */
    Tcl_Channel    tchan_;	 /* place to write trace records */
    
    
    /* Static state. */
    arcp arcp_;
    
    /* Dynamic state. */
    arcv arcv_;		  /* core arc variables */
    TracedInt curq_;	  /* current qlen seen by arrivals */
    int bcount_;	  /* byte count (for qlen) */
    
    /* Statistics Update Timer */
    ARCUpdateTimer update_timer_; 
};

#endif
